import sys
import datetime
from typing import Dict
from mcp.server.fastmcp import FastMCP
from chartmogul_mcp import api_client
from chartmogul_mcp import utils
from chartmogul_mcp.utils import LOGGER
from dotenv import load_dotenv


class ChartMogulMcp:

    def __init__(self):
        load_dotenv()

        # Initialize MCP Server
        self.mcp = FastMCP(utils.MCP_SERVER_NAME, deps=utils.DEPENDENCIES)
        LOGGER.info("ChartMogul MCP Server initialized")

        self.config = api_client.init_chartmogul_config()

        # Register MCP tools
        self._register_tools()


    def _register_tools(self):
        """Register MCP tools to interact with ChartMogul API."""

        ## account
        @self.mcp.tool(name='retrieve_account',
                       description='Retrieve some useful information about your ChartMogul account.')
        async def retrieve_account() -> Dict:
            return api_client.retrieve_account(self.config)

        ## data sources
        @self.mcp.tool(name='list_sources',
                       description='Get a list of all data sources in your ChartMogul account.'
                                   'You can also filter using the data source name or system '
                                   '(the type of system of the data sources, e.g., Stripe, Recurly, Custom, etc.).')
        async def list_sources(name: str = None, system: str = None) -> list:
            return api_client.list_sources(self.config, name, system)

        @self.mcp.tool(name='retrieve_source',
                       description='Retrieve a data source from your ChartMogul account using its UUID.')
        async def retrieve_source(uuid: str) -> Dict:
            return api_client.retrieve_source(self.config, uuid)

        ## customers
        @self.mcp.tool(name='list_customers',
                       description='Get a list of customers in your ChartMogul account. '
                                   'We have a default limit of 20 customers, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.'
                                   'You can also filter based on data_source_uuid, external_id, '
                                   'status (one of New_Lead, Working_Lead, Qualified_Lead, Unqualified_Lead, Active, '
                                   'Past_Due or Cancelled) and system (the type of system of the data sources, '
                                   'e.g. Stripe, Recurly, Custom, etc.).')
        async def list_customers(data_source_uuid: str = None, external_id: str = None, status: str = None,
                                 system: str = None, limit: int = 20) -> list:
            return api_client.list_customers(self.config, data_source_uuid, external_id, status, system, limit)

        @self.mcp.tool(name='search_customers',
                       description='Search a list of all customers with the specified email address '
                                   'in your ChartMogul account.'
                                   'We have a default limit of 20 customers, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.')
        async def search_customers(email: str, limit: int = 20) -> list:
            return api_client.search_customers(self.config, email, limit)

        @self.mcp.tool(name='retrieve_customer',
                       description='Retrieve a customer from your ChartMogul account using its UUID.')
        async def retrieve_customer(uuid: str) -> Dict:
            return api_client.retrieve_customer(self.config, uuid)

        @self.mcp.tool(name='create_customer',
                       description='Create a customer in your ChartMogul account. '
                                   'IMPORTANT: Always ask for ALL missing required details before creating a customer. '
                                   'Required fields are: data_source_uuid, external_id. '
                                   'Optional fields are: company, country (Country code of customer location as per '
                                   'ISO-3166 alpha-2 standard, e.g. US or DE), state (The US state where the customer '
                                   'is located — the US state code as per the ISO-3166-2 standard, e.g. US-CA or '
                                   'US-NY.), city, zip, lead_created_at (ISO 8601-formatted time, in the past), '
                                   'free_trial_started_at (ISO 8601-formatted time, in the past), '
                                   'attributes (an array containing the additional attributes of the customer '
                                   'in the form of tags and custom attributes; tags: An array of tags, '
                                   'custom: An array of custom attributes to be added to the customer. Each custom '
                                   'attribute must have a type, key and value), owner (email address of the user that '
                                   'owns this customer), primary_contact (object which contains first_name, last_name, '
                                   'email, title, phone, linked_in, twitter, notes), website_url.'
                                   'All fields should be included in a data dictionary.')
        async def create_customer(data: dict) -> Dict:
            return api_client.create_customer(self.config, data)

        @self.mcp.tool(name='update_customer',
                       description='Update certain modifiable attributes of a customer in your ChartMogul account. '
                                   'Attributes that can be modified are: company, lead_created_at, '
                                   'free_trial_started_at, zip, city, state, country, attributes '
                                   '(nested object which includes tags and custom object), owner, '
                                   'primary_contact, status and website_url, and should be included in a '
                                   'data dictionary.')
        async def update_customer(uuid: str, data: dict) -> Dict:
            return api_client.update_customer(self.config, uuid, data)

        @self.mcp.tool(name='list_customer_subscriptions',
                       description='Get a list of all subscriptions with the specified customer uuid '
                                   'in your ChartMogul account.'
                                   'We have a default limit of 20 subscriptions, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.')
        async def list_customer_subscriptions(uuid: str, limit: int = 20) -> list:
            return api_client.list_customer_subscriptions(self.config, uuid, limit)

        @self.mcp.tool(name='list_customer_activities',
                       description='Get a list of all activities with the specified customer uuid '
                                   'in your ChartMogul account.'
                                   'We have a default limit of 20 activities, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.')
        async def list_customer_activities(uuid: str, limit: int = 20) -> list:
            return api_client.list_customer_activities(self.config, uuid, limit)

        @self.mcp.tool(name='list_customer_attributes',
                       description='Get a list of all customer attributes with the specified customer uuid '
                                   'in your ChartMogul account.')
        async def list_customer_attributes(uuid: str) -> list:
            return api_client.list_customer_attributes(self.config, uuid)

        @self.mcp.tool(name='add_customer_tags',
                       description='Add a list of tags to the specified customer uuid in your ChartMogul account.'
                                   'This endpoint is idempotent, meaning that calling it multiple times with the same '
                                   'tags, no additional tags will be added to the customer. If the endpoint is '
                                   'called with a single new tag, it will be added to previous tags of the user.')
        async def add_customer_tags(uuid: str, tags: list) -> list:
            return api_client.add_customer_tags(self.config, uuid, tags)

        @self.mcp.tool(name='add_customer_custom_attributes',
                       description='Add a list of custom attributes to the specified customer uuid in your ChartMogul account.'
                                   'custom_attributes is an array containing the custom attributes to be added to the '
                                   'customer. Each custom attribute must have a type, key and value as described below:'
                                   'type: The data type of the custom attribute. Can be one of String, Integer, Decimal, '
                                   'Timestamp or Boolean. key: The name of the custom attribute. Accepts alphanumeric '
                                   'characters and underscores, value: The value of the custom attribute. '
                                   'Should be of the data type as specified in type. source: Optional parameter to '
                                   'specify where the custom attribute was created, displayed in the ChartMogul UI. '
                                   'Defaults to API.')
        async def add_customer_custom_attributes(uuid: str, custom_attributes: list) -> list:
            return api_client.add_customer_custom_attributes(self.config, uuid, custom_attributes)

        ## contacts
        @self.mcp.tool(name='list_contacts',
                       description='Get a list of all contacts in your ChartMogul account.'
                                   'We have a default limit of 20 contacts, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.'
                                   'You can filter using the contact email address and the customer_external_id.')
        async def list_contacts(email: str = None, customer_external_id: str = None, limit: int = 20) -> list:
            return api_client.list_contacts(self.config, email, customer_external_id, limit)

        @self.mcp.tool(name='retrieve_contact',
                       description='Retrieve a contact from your ChartMogul account using its UUID.')
        async def retrieve_contact(uuid: str) -> Dict:
            return api_client.retrieve_contact(self.config, uuid)

        @self.mcp.tool(name='update_contact',
                       description='Update certain modifiable attributes of a contact in your ChartMogul account. '
                                   'Attributes that can be modified are: first_name, last_name, '
                                   'position, title, email, phone, linked_in, twitter, notes, custom '
                                   '(an array containing the custom attributes; each custom attribute must be defined '
                                   'as an object with a key and value), and should be included in a '
                                   'data dictionary.')
        async def update_contact(uuid: str, data: dict) -> Dict:
            return api_client.update_contact(self.config, uuid, data)

        @self.mcp.tool(name='create_contact',
                       description='Create a contact in your ChartMogul account. '
                                   'IMPORTANT: Always ask for ALL missing required details before creating a contact. '
                                   'Required fields are: customer_uuid, data_source_uuid. '
                                   'Optional fields are: first_name, last_name, position, title, email, phone, linked_in, '
                                   'twitter, notes, custom (an array containing the custom attributes; each '
                                   'custom attribute must be defined as an object with a key and value). '
                                   'All fields should be included in a data dictionary.')
        async def create_contact(data: dict) -> Dict:
            return api_client.create_contact(self.config, data)


        ## customer_notes
        @self.mcp.tool(name='list_customer_notes',
                       description='Get a list of all customer notes in your ChartMogul account.'
                                   'We have a default limit of 20 customer notes, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.'
                                   'You can filter using the customer_uuid and the type (note or call).')
        async def list_customer_notes(customer_uuid: str = None, type: str = None, limit: int = 20) -> list:
            return api_client.list_customer_notes(self.config, customer_uuid, type, limit)

        @self.mcp.tool(name='retrieve_customer_note',
                       description='Retrieve a customer note from your ChartMogul account using its UUID.')
        async def retrieve_customer_note(uuid: str) -> Dict:
            return api_client.retrieve_customer_note(self.config, uuid)

        @self.mcp.tool(name='update_customer_note',
                       description='Update certain modifiable attributes of a customer note in your ChartMogul account. '
                                   'Attributes that can be modified are: author_email, text, '
                                   'call_duration (integer; relevant for type call; duration of the call in seconds), '
                                   'created_at (an ISO 8601-formatted time in the past), '
                                   'updated_at (an ISO 8601-formatted time in the past), and should be included in a '
                                   'data dictionary.')
        async def update_customer_note(uuid: str, data: dict) -> Dict:
            return api_client.update_customer_note(self.config, uuid, data)

        @self.mcp.tool(name='create_customer_note',
                       description='Create a customer note in your ChartMogul account. '
                                   'IMPORTANT: Always ask for ALL missing required details before creating a customer note. '
                                   'Required fields are: customer_uuid, type (call or note). '
                                   'Optional fields are: author_email, text, '
                                   'call_duration (integer; relevant for type call; duration of the call in seconds), '
                                   'created_at (an ISO 8601-formatted time in the past). '
                                   'All fields should be included in a data dictionary.')
        async def create_customer_note(data: dict) -> Dict:
            return api_client.create_customer_note(self.config, data)


        ## opportunities
        @self.mcp.tool(name='list_opportunities',
                       description='Get a list of all opportunities in your ChartMogul account.'
                                   'We have a default limit of 20 opportunities, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.'
                                   'You can filter using the customer_uuid, owner (email address of the ChartMogul user '
                                   'with a CRM seat who is the primary salesperson responsible for this opportunity), '
                                   'pipeline, pipeline_stage, estimated_close_date_on_or_after '
                                   '(lower limit of the estimated close date range; an ISO 8601-formatted date) and '
                                   'estimated_close_date_on_or_before (upper limit of the estimated close date range; '
                                   'an ISO 8601-formatted date).')
        async def list_opportunities(customer_uuid: str = None, owner: str = None, pipeline: str = None,
                                     pipeline_stage: str = None,
                                     estimated_close_date_on_or_after: datetime.datetime =None,
                                     estimated_close_date_on_or_before: datetime.datetime =None,
                                     limit: int = 20) -> list:
            return api_client.list_opportunities(self.config, customer_uuid, owner, pipeline, pipeline_stage,
                                                 estimated_close_date_on_or_after, estimated_close_date_on_or_before,
                                                 limit)

        @self.mcp.tool(name='retrieve_opportunity',
                       description='Retrieve an opportunity from your ChartMogul account using its UUID.')
        async def retrieve_opportunity(uuid: str) -> Dict:
            return api_client.retrieve_opportunity(self.config, uuid)

        @self.mcp.tool(name='update_opportunity',
                       description='Update certain modifiable attributes of an opportunity in your ChartMogul account. '
                                   'Attributes that can be modified are: owner, pipeline, pipeline_stage, '
                                   'estimated_close_date, amount_in_cents, currency (The 3-letter currency code for '
                                   'the expected close value, e.g. USD, EUR or GBP), type (recurring or one-time), '
                                   'forecast_category (pipeline, best_case, committed, lost or won), win_likelihood '
                                   '(0-100), custom (list of custom attributes as key and value pairs) '
                                   'and should be included in a data dictionary.')
        async def update_opportunity(uuid: str, data: dict) -> Dict:
            return api_client.update_opportunity(self.config, uuid, data)

        @self.mcp.tool(name='create_opportunity',
                       description='Create an opportunity in your ChartMogul account. '
                                   'IMPORTANT: Always ask for ALL missing required details before creating an opportunity. '
                                   'Required fields are: customer_uuid, owner '
                                   '(email address of the ChartMogul user with a CRM seat who is the primary '
                                   'salesperson responsible for this opportunity), pipeline, '
                                   'pipeline_stage, estimated_close_date (an ISO 8601-formatted date), '
                                   'amount_in_cents, currency (The 3-letter currency code for the expected close value, '
                                   'e.g. USD, EUR or GBP). '
                                   'Optional fields: type (recurring or one-time), forecast_category (pipeline, '
                                   'best_case, committed, lost or won), win_likelihood (integer; 0-100), '
                                   'custom (list of custom attributes as key and value pairs). '
                                   'All fields should be included in a data dictionary.')
        async def create_opportunity(data: dict) -> Dict:
            return api_client.create_opportunity(self.config, data)


        ## plans
        @self.mcp.tool(name='list_plans',
                       description='Get a list of all plans in your ChartMogul account.'
                                   'We have a default limit of 20 plans, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.'
                                   'You can filter using the data_source_uuid, external_id, and system (the billing system '
                                   'that the plan belongs to, e.g., Stripe, Recurly, Custom).')
        async def list_plans(data_source_uuid: str = None, external_id: str = None, system: str = None,
                             limit: int = 20) -> list:
            return api_client.list_plans(self.config, data_source_uuid, external_id, system, limit)

        @self.mcp.tool(name='retrieve_plan',
                       description='Retrieve a plan from your ChartMogul account using its UUID.')
        async def retrieve_plan(uuid: str) -> Dict:
            return api_client.retrieve_plan(self.config, uuid)

        @self.mcp.tool(name='update_plan',
                       description='Update certain modifiable attributes of a plan in your ChartMogul account. '
                                   'Attributes that can be modified are: name, interval_count '
                                   '(frequency of billing interval; accepts integers greater than 0, '
                                   'e.g., 6 for a half-yearly plan), interval_unit (day, month or year) '
                                   'and should be included in a data dictionary.')
        async def update_plan(uuid: str, data: dict) -> Dict:
            return api_client.update_plan(self.config, uuid, data)

        @self.mcp.tool(name='create_plan',
                       description='Create a plan in your ChartMogul account. '
                                   'IMPORTANT: Always ask for ALL missing required details before creating a plan. '
                                   'Required fields are: data_source_uuid, name, '
                                   'interval_count (frequency of billing interval; accepts integers greater than 0, '
                                   'e.g., 6 for a half-yearly plan), interval_unit (day, month or year). '
                                   'Optional field: external_id. '
                                   'All fields should be included in a data dictionary.')
        async def create_plan(data: dict) -> Dict:
            return api_client.create_plan(self.config, data)


        ## plan groups
        @self.mcp.tool(name='list_plan_groups',
                       description='Get a list of all plan groups in your ChartMogul account.'
                                   'We have a default limit of 20 plan groups, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.')
        async def list_plan_groups(limit: int = 20) -> list:
            return api_client.list_plan_groups(self.config, limit)

        ## plan groups
        @self.mcp.tool(name='list_plan_group_plans',
                       description='Get a list of all plans in a plan group using its UUID.')
        async def list_plan_group_plans(uuid: str = None, limit: int = 20) -> list:
            return api_client.list_plan_group_plans(self.config, uuid, limit)

        @self.mcp.tool(name='retrieve_plan_group',
                       description='Retrieve a plan group from your ChartMogul account using its UUID.')
        async def retrieve_plan_group(uuid: str) -> Dict:
            return api_client.retrieve_plan_group(self.config, uuid)

        @self.mcp.tool(name='update_plan_group',
                       description='Update certain modifiable attributes of a plan group in your ChartMogul account. '
                                   'Attributes that can be modified are: name, plans '
                                   '(array of the uuids of the plans to be added to the plan group) '
                                   'and should be included in a data dictionary.')
        async def update_plan_group(uuid: str, data: dict) -> Dict:
            return api_client.update_plan_group(self.config, uuid, data)

        @self.mcp.tool(name='create_plan_group',
                       description='Create a plan group in your ChartMogul account. '
                                   'IMPORTANT: Always ask for ALL missing required details before creating a plan group. '
                                   'Required fields are: name, plans '
                                   '(array of the uuids of the plans to be added to the plan group) '
                                   'and should be included in a data dictionary.')
        async def create_plan_group(data: dict) -> Dict:
            return api_client.create_plan_group(self.config, data)


        ## tasks
        @self.mcp.tool(name='list_tasks',
                       description='Get a list of all tasks in your ChartMogul account.'
                                   'We have a default limit of 20 tasks, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.'
                                   'You can filter using the customer_uuid, assignee (email address of the ChartMogul user '
                                   'with a CRM seat assigned to the task), due_date_on_or_after '
                                   '(lower limit of the due date range; an ISO 8601-formatted date), '
                                   'estimated_close_date_on_or_before (upper limit of the due date range; '
                                   'an ISO 8601-formatted date), completed (true or false).'
                       )
        async def list_tasks(customer_uuid: str = None, assignee: str = None,
                             due_date_on_or_after: datetime.datetime = None,
                             estimated_close_date_on_or_before: datetime.datetime = None, completed: bool = None,
                             limit: int = 20) -> list:
            return api_client.list_tasks(self.config, customer_uuid, assignee, due_date_on_or_after,
                                         estimated_close_date_on_or_before, completed, limit)

        @self.mcp.tool(name='retrieve_task',
                       description='Retrieve a task from your ChartMogul account using its UUID.')
        async def retrieve_task(uuid: str) -> Dict:
            return api_client.retrieve_task(self.config, uuid)

        @self.mcp.tool(name='update_task',
                       description='Update certain modifiable attributes of a task in your ChartMogul account. '
                                   'Attributes that can be modified are: task_details '
                                   '(up to 255 characters), assignee  (email address of the ChartMogul user '
                                   'with a CRM seat assigned to the task), due_date (an ISO 8601-formatted date), '
                                   'completed_at (an ISO 8601-formatted date) '
                                   'and should be included in a data dictionary.')
        async def update_task(uuid: str, data: dict) -> Dict:
            return api_client.update_task(self.config, uuid, data)

        @self.mcp.tool(name='create_task',
                       description='Create a task in your ChartMogul account. '
                                   'IMPORTANT: Always ask for ALL missing required details before creating a task. '
                                   'Required fields are: customer_uuid, task_details '
                                   '(up to 255 characters), assignee (email address of the ChartMogul user '
                                   'with a CRM seat assigned to the task), due_date (an ISO 8601-formatted date). '
                                   'Optional field: completed_at (an ISO 8601-formatted date).'
                                   'All fields should be included in a data dictionary.')
        async def create_task(data: dict) -> Dict:
            return api_client.create_task(self.config, data)

        ## metrics api
        @self.mcp.tool(name='all_metrics',
                       description='Retrieve all key metrics, for the specified time period, interval and filters. '
                                   'Metrics include: MRR, ARR, ARPA, ASP, customer count, customer churn rate, '
                                   'MRR churn rate and LTV. '
                                   'Provide the start-date, end-date and interval (possible values are day, week, month,'
                                   'quarter or year. Additional filter values include geo (A comma-separated list of ISO '
                                   '3166-1 Alpha-2 formatted country codes) and plans (A comma-separated list of '
                                   'plan names (as configured in your ChartMogul account), UUIDs and external IDs to '
                                   'filter the results. Spaces in plan names must be URL-encoded, '
                                   'e.g., Silver%20plan,Gold%20plan,Enterprise%20plan).'
                                   'All amounts are given in the selected currency of your account '
                                   'and are an integer number of cents. Divide by 100 to obtain the actual value.')
        async def all_metrics(start_date: str, end_date: str, interval: str, geo: str = None,
                              plans: str = None) -> list:
            return api_client.all_metrics(self.config, start_date, end_date, interval, geo, plans)

        @self.mcp.tool(name='mrr_metrics',
                       description='Retrieve Monthly Recurring Revenue (MRR) metrics, for the specified time period, interval and filters. '
                                   'Provide the start-date, end-date and interval (possible values are day, week, month,'
                                   'quarter or year. Additional filter values include geo (A comma-separated list of ISO '
                                   '3166-1 Alpha-2 formatted country codes) and plans (A comma-separated list of '
                                   'plan names (as configured in your ChartMogul account), UUIDs and external IDs to '
                                   'filter the results. Spaces in plan names must be URL-encoded, '
                                   'e.g., Silver%20plan,Gold%20plan,Enterprise%20plan).'
                                   'All amounts are given in the selected currency of your account '
                                   'and are an integer number of cents. Divide by 100 to obtain the actual value.')
        async def mrr_metrics(start_date: str, end_date: str, interval: str, geo: str = None,
                              plans: str = None) -> list:
            return api_client.mrr_metrics(self.config, start_date, end_date, interval, geo, plans)

        @self.mcp.tool(name='arr_metrics',
                       description='Retrieve Annualized Run Rate (ARR) metrics, for the specified time period, interval and filters. '
                                   'Provide the start-date, end-date and interval (possible values are day, week, month,'
                                   'quarter or year. Additional filter values include geo (A comma-separated list of ISO '
                                   '3166-1 Alpha-2 formatted country codes) and plans (A comma-separated list of '
                                   'plan names (as configured in your ChartMogul account), UUIDs and external IDs to '
                                   'filter the results. Spaces in plan names must be URL-encoded, '
                                   'e.g., Silver%20plan,Gold%20plan,Enterprise%20plan).'
                                   'All amounts are given in the selected currency of your account '
                                   'and are an integer number of cents. Divide by 100 to obtain the actual value.')
        async def arr_metrics(start_date: str, end_date: str, interval: str, geo: str = None,
                              plans: str = None) -> list:
            return api_client.arr_metrics(self.config, start_date, end_date, interval, geo, plans)

        @self.mcp.tool(name='arpa_metrics',
                       description='Retrieve Average Revenue Per Account (ARPA) metrics, for the specified time period, interval and filters. '
                                   'Provide the start-date, end-date and interval (possible values are day, week, month,'
                                   'quarter or year. Additional filter values include geo (A comma-separated list of ISO '
                                   '3166-1 Alpha-2 formatted country codes) and plans (A comma-separated list of '
                                   'plan names (as configured in your ChartMogul account), UUIDs and external IDs to '
                                   'filter the results. Spaces in plan names must be URL-encoded, '
                                   'e.g., Silver%20plan,Gold%20plan,Enterprise%20plan).'
                                   'All amounts are given in the selected currency of your account '
                                   'and are an integer number of cents. Divide by 100 to obtain the actual value.')
        async def arpa_metrics(start_date: str, end_date: str, interval: str, geo: str = None,
                               plans: str = None) -> list:
            return api_client.arpa_metrics(self.config, start_date, end_date, interval, geo, plans)

        @self.mcp.tool(name='asp_metrics',
                       description='Retrieve Average Sale Price (ASP) metrics, for the specified time period, interval and filters. '
                                   'Provide the start-date, end-date and interval (possible values are month,'
                                   'quarter or year. Additional filter values include geo (A comma-separated list of ISO '
                                   '3166-1 Alpha-2 formatted country codes) and plans (A comma-separated list of '
                                   'plan names (as configured in your ChartMogul account), UUIDs and external IDs to '
                                   'filter the results. Spaces in plan names must be URL-encoded, '
                                   'e.g., Silver%20plan,Gold%20plan,Enterprise%20plan).'
                                   'All amounts are given in the selected currency of your account '
                                   'and are an integer number of cents. Divide by 100 to obtain the actual value.')
        async def asp_metrics(start_date: str, end_date: str, interval: str, geo: str = None,
                              plans: str = None) -> list:
            return api_client.asp_metrics(self.config, start_date, end_date, interval, geo, plans)

        @self.mcp.tool(name='customer_count_metrics',
                       description='Retrieve customer count metrics, for the specified time period, interval and filters. '
                                   'Provide the start-date, end-date and interval (possible values are day, week, month,'
                                   'quarter or year. Additional filter values include geo (A comma-separated list of ISO '
                                   '3166-1 Alpha-2 formatted country codes) and plans (A comma-separated list of '
                                   'plan names (as configured in your ChartMogul account), UUIDs and external IDs to '
                                   'filter the results. Spaces in plan names must be URL-encoded, '
                                   'e.g., Silver%20plan,Gold%20plan,Enterprise%20plan).')
        async def customer_count_metrics(start_date: str, end_date: str, interval: str, geo: str = None,
                                         plans: str = None) -> list:
            return api_client.customer_count_metrics(self.config, start_date, end_date, interval, geo, plans)

        @self.mcp.tool(name='customer_churn_rate_metrics',
                       description='Retrieve customer churn rate metrics, for the specified time period, interval and filters. '
                                   'Provide the start-date, end-date and interval (possible values are day, week, month,'
                                   'quarter or year. Additional filter values include geo (A comma-separated list of ISO '
                                   '3166-1 Alpha-2 formatted country codes) and plans (A comma-separated list of '
                                   'plan names (as configured in your ChartMogul account), UUIDs and external IDs to '
                                   'filter the results. Spaces in plan names must be URL-encoded, '
                                   'e.g., Silver%20plan,Gold%20plan,Enterprise%20plan).')
        async def customer_churn_rate_metrics(start_date: str, end_date: str, interval: str, geo: str = None,
                                              plans: str = None) -> list:
            return api_client.customer_churn_rate_metrics(self.config, start_date, end_date, interval, geo, plans)

        @self.mcp.tool(name='mrr_churn_rate_metrics',
                       description='Retrieve Net MRR Churn Rate metrics, for the specified time period, interval and filters. '
                                   'Provide the start-date, end-date and interval (possible values are day, week, month,'
                                   'quarter or year. Additional filter values include geo (A comma-separated list of ISO '
                                   '3166-1 Alpha-2 formatted country codes) and plans (A comma-separated list of '
                                   'plan names (as configured in your ChartMogul account), UUIDs and external IDs to '
                                   'filter the results. Spaces in plan names must be URL-encoded, '
                                   'e.g., Silver%20plan,Gold%20plan,Enterprise%20plan).')
        async def mrr_churn_rate_metrics(start_date: str, end_date: str, interval: str, geo: str = None,
                                         plans: str = None) -> list:
            return api_client.mrr_churn_rate_metrics(self.config, start_date, end_date, interval, geo, plans)

        @self.mcp.tool(name='ltv_metrics',
                       description='Retrieve Customer Lifetime Value (LTV) metrics, for the specified time period, and filters. '
                                   'Provide the start-date, end-date. Additional filter values include geo (A comma-separated list of ISO '
                                   '3166-1 Alpha-2 formatted country codes) and plans (A comma-separated list of '
                                   'plan names (as configured in your ChartMogul account), UUIDs and external IDs to '
                                   'filter the results. Spaces in plan names must be URL-encoded, '
                                   'e.g., Silver%20plan,Gold%20plan,Enterprise%20plan).'
                                   'All amounts are given in the selected currency of your account '
                                   'and are an integer number of cents. Divide by 100 to obtain the actual value.')
        async def ltv_metrics(start_date: str, end_date: str, interval: str, geo: str = None,
                              plans: str = None) -> list:
            return api_client.ltv_metrics(self.config, start_date, end_date, interval, geo, plans)

        ## subscription events
        @self.mcp.tool(name='list_subscription_events',
                       description='Get a list of all subscription events in your ChartMogul account.'
                                   'We have a default limit of 20 subscription events, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.'
                                   'You can filter using the data_source_uuid, external_id of a subscription event, '
                                   'customer_external_id, subscription_external_id, event_type (one of '
                                   'subscription_start, subscription_start_scheduled, '
                                   'scheduled_subscription_start_retracted, subscription_cancelled, '
                                   'subscription_cancellation_scheduled, scheduled_subscription_cancellation_retracted, '
                                   'subscription_updated, subscription_update_scheduled, '
                                   'scheduled_subscription_update_retracted, subscription_event_retracted), event_date '
                                   '(an ISO 8601 formatted time), effective_date (an ISO 8601 formatted time), '
                                   'plan_external_id.')
        async def list_subscription_events(data_source_uuid: str = None, external_id: str = None, customer_external_id: str = None,
                                           subscription_external_id: str = None, event_type: str = None,
                                           event_date: datetime.datetime = None,
                                           effective_date: datetime.datetime = None, plan_external_id: str = None,
                                           limit: int = 20) -> list:
            return api_client.list_subscription_events(self.config, data_source_uuid, external_id, customer_external_id,
                                                       subscription_external_id, event_type, event_date, effective_date,
                                                       plan_external_id, limit)

        ## invoices
        @self.mcp.tool(name='list_invoices',
                       description='Get a list of all invoices in your ChartMogul account.'
                                   'We have a default limit of 20 invoices, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.'
                                   'You can filter using the data_source_uuid, invoice external_id, customer_uuid and '
                                   'validation_type (one of valid, invalid or all).')
        async def list_invoices(data_source_uuid: str = None, external_id: str = None, customer_uuid: str = None,
                                validation_type: str = None, limit: int = 20) -> list:
            return api_client.list_invoices(self.config, data_source_uuid, external_id, customer_uuid, validation_type,
                                            limit)

        ## activities
        @self.mcp.tool(name='list_activities',
                       description='Get a list of all activities in your ChartMogul account.'
                                   'We have a default limit of 20 activities, '
                                   'ask but discourage the user if they want more than 20 as this will exhaust AI tokens.'
                                   'You can filter with start_date (an ISO 8601 formatted time), end_date '
                                   '(an ISO 8601 formatted time), type (one of new_biz, reactivation, expansion, '
                                   'contraction or churn) and order (Setting value as -date returns results in '
                                   'descending order with the latest activity returned first, while date returns '
                                   'results in ascending order.). ')
        async def list_activities(start_date: datetime.datetime = None, end_date: datetime.datetime = None,
                                  type: str = None, order: str = None, limit: int = 20) -> list:
            return api_client.list_activities(self.config, start_date, end_date, type, order, limit)


    def run(self):
        """Start the MCP server."""
        try:
            LOGGER.info("Running MCP Server for ChartMogul API interactions")
            self.mcp.run(transport="stdio")
        except Exception as e:
            LOGGER.error(f"Fatal Error in ChartMogul MCP Server: {str(e)}", exc_info=True)
            sys.exit(1)



import os
from dotenv import load_dotenv
import logging

load_dotenv()

CHARTMOGUL_TOKEN = os.getenv('CHARTMOGUL_TOKEN')
MCP_SERVER_NAME = "mcp-chartmogul"
DEPENDENCIES = [
    "chartmogul",
    "python-dotenv"
]

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)
LOGGER = logging.getLogger(MCP_SERVER_NAME)


import datetime
import chartmogul
from chartmogul_mcp import utils
from chartmogul_mcp.utils import LOGGER


def init_chartmogul_config():
    return chartmogul.Config(utils.CHARTMOGUL_TOKEN)

## Account Endpoint

def retrieve_account(config):
    """
    Retrieve the account information.

    """
    LOGGER.info(f"Retrieve account information.")
    request = chartmogul.Account.retrieve(config)
    try:
        account = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error retrieving customer: {str(e)}", exc_info=True)
        return None
    return account


## Data sources Endpoints

def list_sources(config, name=None, system=None):
    """
    List all data sources from ChartMogul API.

    Returns: A list of ChartMogul data sources.
    """
    LOGGER.info(f"List data sources {name}, {system}.")
    all_sources = []
    request = chartmogul.DataSource.all(config, name=name, system=system)
    try:
        sources = request.get()
        all_sources.extend([parse_object(entry) for entry in sources.data_sources])
    except Exception as e:
        LOGGER.error(f"Error listing data sources: {str(e)}", exc_info=True)
        return None
    return all_sources


def retrieve_source(config, data_source_uuid):
    """
    Retrieve a data source from ChartMogul API.

    Returns: The data source.
    """
    LOGGER.info(f"Retrieve data source for {data_source_uuid}.")
    request = chartmogul.DataSource.retrieve(config, uuid=data_source_uuid)
    try:
        source = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error retrieving data source: {str(e)}", exc_info=True)
        return None
    return source


## Customers Endpoints

def list_customers(config, data_source_uuid=None, external_id=None, status=None, system=None, limit=20) -> list:
    """
    List all customers from ChartMogul API.
        
    Returns: A list of ChartMogul customers.
    """
    LOGGER.info(f"List customers for {data_source_uuid}, {external_id}, {status}, {system}.")
    all_customers = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.Customer.all(config,
                                          data_source_uuid=data_source_uuid,
                                          external_id=external_id,
                                          status=status,
                                          system=system,
                                          cursor=cursor,
                                          per_page=per_page)
        try:
            customers = request.get()
            all_customers.extend([parse_object(entry) for entry in customers.entries])
            total += per_page
            has_more = customers.has_more
            cursor = customers.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul customers: {str(e)}", exc_info=True)
            return None
    return all_customers


def create_customer(config, data):
    """
    Create a customer from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Creating customer {data}.")
    request = chartmogul.Customer.create(config, data=data)
    try:
        customer = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error creating customer: {str(e)}", exc_info=True)
        return None
    return customer

def retrieve_customer(config, uuid):
    """
    Retrieve a customer from ChartMogul API.

    Returns: The customer.
    """
    LOGGER.info(f"Retrieving customer for {uuid}.")
    request = chartmogul.Customer.retrieve(config, uuid=uuid)
    try:
        customer = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error retrieving customer: {str(e)}", exc_info=True)
        return None
    return customer


def update_customer(config, uuid, data):
    """
    Update a customer from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Updating customer {uuid}, {data}.")
    request = chartmogul.Customer.modify(config, uuid=uuid, data=data)
    try:
        customer = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error updating customer: {str(e)}", exc_info=True)
        return None
    return customer


def search_customers(config, email, limit=20) -> list:
    """
    Search all customers by email from ChartMogul API.

    Returns: A list of ChartMogul customers.
    """
    LOGGER.info(f"Search customers for {email}.")
    all_customers = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.Customer.search(config, email=email, cursor=cursor, per_page=per_page)
        try:
            customers = request.get()
            all_customers.extend([parse_object(entry) for entry in customers.entries])
            total += per_page
            has_more = customers.has_more
            cursor = customers.cursor
        except Exception as e:
            LOGGER.error(f"Error searching ChartMogul customers: {str(e)}", exc_info=True)
            return None
    return all_customers


def list_customer_subscriptions(config, uuid=None, limit=20) -> list:
    """
    List all subscriptions of a customer from ChartMogul API.

    Returns: A list of ChartMogul subscriptions.
    """
    LOGGER.info(f"List subscriptions for {uuid}.")
    all_subscriptions = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.CustomerSubscription.all(config, uuid=uuid, cursor=cursor, per_page=per_page)
        try:
            subscriptions = request.get()
            all_subscriptions.extend([parse_object(entry) for entry in subscriptions.entries])
            total += per_page
            has_more = subscriptions.has_more
            cursor = subscriptions.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul subscriptions: {str(e)}", exc_info=True)
            return None
    return all_subscriptions


def list_customer_activities(config, uuid=None, limit=20) -> list:
    """
    List all activities of a customer from ChartMogul API.

    Returns: A list of ChartMogul activities.
    """
    LOGGER.info(f"List activities for {uuid}.")
    all_activities = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.CustomerActivity.all(config, uuid=uuid, cursor=cursor, per_page=per_page)
        try:
            activities = request.get()
            all_activities.extend([parse_object(entry) for entry in activities.entries])
            total += per_page
            has_more = activities.has_more
            cursor = activities.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul activities: {str(e)}", exc_info=True)
            return None
    return all_activities


def list_customer_attributes(config, uuid) -> list:
    """
    List all attributes of a customer from ChartMogul API.

    Returns: A list of ChartMogul attributes.
    """
    LOGGER.info(f"List attributes for {uuid}.")
    request = chartmogul.Attributes.retrieve(config, uuid=uuid)
    try:
        attributes = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error fetching attributes: {str(e)}", exc_info=True)
        return None
    return attributes


def add_customer_tags(config, uuid, data) -> list:
    """
    Add tags to customer using ChartMogul API.

    Returns: A list of ChartMogul tags added.
    """
    LOGGER.info(f"Add tags for {uuid}, {data}.")
    request = chartmogul.Tags.add(config, uuid=uuid, data={"tags": data})
    try:
        tags = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error adding tags: {str(e)}", exc_info=True)
        return None
    return tags


def add_customer_custom_attributes(config, uuid, data) -> list:
    """
    Add custom attributes to customer using ChartMogul API.

    Returns: A list of ChartMogul custom attributes added.
    """
    LOGGER.info(f"Add custom attributes for {uuid}, {data}.")
    request = chartmogul.CustomAttributes.add(config, uuid=uuid, data={"custom": data})
    try:
        custom_attributes = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error adding custom attributes: {str(e)}", exc_info=True)
        return None
    return custom_attributes


## Contacts Endpoints

def list_contacts(config, email=None, customer_external_id=None, limit=20) -> list:
    """
    List all contacts from ChartMogul API.

    Returns: A list of ChartMogul contacts.
    """
    LOGGER.info(f"List contacts for {email}, {customer_external_id}.")
    all_contacts = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.Contact.all(config,
                                         email=email,
                                         customer_external_id=customer_external_id,
                                         cursor=cursor,
                                         per_page=per_page)
        try:
            contacts = request.get()
            all_contacts.extend([parse_object(entry) for entry in contacts.entries])
            total += per_page
            has_more = contacts.has_more
            cursor = contacts.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul contacts: {str(e)}", exc_info=True)
            return None
    return all_contacts


def retrieve_contact(config, uuid):
    """
    Retrieve a contact from ChartMogul API.

    Returns: The contact.
    """
    LOGGER.info(f"Retrieving contact for {uuid}.")
    request = chartmogul.Contact.retrieve(config, uuid=uuid)
    try:
        contact = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error retrieving contact: {str(e)}", exc_info=True)
        return None
    return contact


def create_contact(config, data):
    """
    Create a contact from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Creating contact {data}.")
    request = chartmogul.Contact.create(config, data=data)
    try:
        contact = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error creating contact: {str(e)}", exc_info=True)
        return None
    return contact


def update_contact(config, uuid, data):
    """
    Update a contact from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Updating contact {uuid}, {data}.")
    request = chartmogul.Contact.modify(config, uuid=uuid, data=data)
    try:
        contact = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error updating contact: {str(e)}", exc_info=True)
        return None
    return contact


## Notes and call logs Endpoints

def list_customer_notes(config, customer_uuid=None, type=None, author_email=None, limit=20) -> list:
    """
    List all customer_notes from ChartMogul API.

    Returns: A list of ChartMogul customer_notes.
    """
    LOGGER.info(f"List customer_notes for {customer_uuid}, {type}, {author_email}.")
    all_customer_notes = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.CustomerNote.all(config,
                                              customer_uuid=customer_uuid,
                                              author_email=author_email,
                                              type=type,
                                              cursor=cursor,
                                              per_page=per_page)
        try:
            customer_notes = request.get()
            all_customer_notes.extend([parse_object(entry) for entry in customer_notes.entries])
            total += per_page
            has_more = customer_notes.has_more
            cursor = customer_notes.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul customer_notes: {str(e)}", exc_info=True)
            return None
    return all_customer_notes


def retrieve_customer_note(config, uuid):
    """
    Retrieve a customer_note from ChartMogul API.

    Returns: The customer_note.
    """
    LOGGER.info(f"Retrieving customer_note for {uuid}.")
    request = chartmogul.CustomerNote.retrieve(config, uuid=uuid)
    try:
        customer_note = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error retrieving customer_note: {str(e)}", exc_info=True)
        return None
    return customer_note


def create_customer_note(config, data):
    """
    Create a customer_note from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Creating contact {data}.")
    request = chartmogul.CustomerNote.create(config, data=data)
    try:
        customer_note = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error creating contact: {str(e)}", exc_info=True)
        return None
    return customer_note


def update_customer_note(config, uuid, data):
    """
    Update a customer_note from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Updating customer_note {uuid}, {data}.")
    request = chartmogul.CustomerNote.patch(config, uuid=uuid, data=data)
    try:
        customer_note = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error updating customer_note: {str(e)}", exc_info=True)
        return None
    return customer_note


## Opportunities Endpoints

def list_opportunities(config, customer_uuid=None, owner=None, pipeline=None, pipeline_stage=None,
                       estimated_close_date_on_or_after=None, estimated_close_date_on_or_before=None,
                       limit=20) -> list:
    """
    List all opportunities from ChartMogul API.

    Returns: A list of ChartMogul opportunities.
    """
    LOGGER.info(f"List opportunities for {customer_uuid}, {owner}, {pipeline}, {pipeline_stage}, "
          f"{estimated_close_date_on_or_after}, {estimated_close_date_on_or_before}.")
    all_opportunities = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.Opportunity.all(config,
                                             customer_uuid=customer_uuid,
                                             owner=owner,
                                             pipeline=pipeline,
                                             pipeline_stage=pipeline_stage,
                                             estimated_close_date_on_or_after=estimated_close_date_on_or_after,
                                             estimated_close_date_on_or_before=estimated_close_date_on_or_before,
                                             cursor=cursor,
                                             per_page=per_page)
        try:
            opportunities = request.get()
            all_opportunities.extend([parse_object(entry) for entry in opportunities.entries])
            total += per_page
            has_more = opportunities.has_more
            cursor = opportunities.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul opportunities: {str(e)}", exc_info=True)
            return None
    return all_opportunities


def retrieve_opportunity(config, uuid):
    """
    Retrieve a opportunity from ChartMogul API.

    Returns: The opportunity.
    """
    LOGGER.info(f"Retrieving opportunity for {uuid}.")
    request = chartmogul.Opportunity.retrieve(config, uuid=uuid)
    try:
        opportunity = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error retrieving opportunity: {str(e)}", exc_info=True)
        return None
    return opportunity


def create_opportunity(config, data):
    """
    Create a opportunity from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Creating opportunity {data}.")
    request = chartmogul.Opportunity.create(config, data=data)
    try:
        opportunity = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error creating opportunity: {str(e)}", exc_info=True)
        return None
    return opportunity


def update_opportunity(config, uuid, data):
    """
    Update a opportunity from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Updating opportunity {uuid}, {data}.")
    request = chartmogul.Opportunity.patch(config, uuid=uuid, data=data)
    try:
        opportunity = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error updating opportunity: {str(e)}", exc_info=True)
        return None
    return opportunity


## Plans Endpoints

def list_plans(config, data_source_uuid=None, external_id=None, system=None, limit=20) -> list:
    """
    List all plans from ChartMogul API.

    Returns: A list of ChartMogul plans.
    """
    LOGGER.info(f"List plans for {data_source_uuid}, {external_id}, {system}.")
    all_plans = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.Plan.all(config,
                                      data_source_uuid=data_source_uuid,
                                      external_id=external_id,
                                      system=system,
                                      cursor=cursor,
                                      per_page=per_page)
        try:
            plans = request.get()
            all_plans.extend([parse_object(entry) for entry in plans.plans])
            total += per_page
            has_more = plans.has_more
            cursor = plans.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul plans: {str(e)}", exc_info=True)
            return None
    return all_plans


def retrieve_plan(config, uuid):
    """
    Retrieve a plan from ChartMogul API.

    Returns: The plan.
    """
    LOGGER.info(f"Retrieving plan for {uuid}.")
    request = chartmogul.Plan.retrieve(config, uuid=uuid)
    try:
        plan = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error retrieving plan: {str(e)}", exc_info=True)
        return None
    return plan


def create_plan(config, data):
    """
    Create a plan from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Creating plan {data}.")
    request = chartmogul.Plan.create(config, data=data)
    try:
        plan = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error creating plan: {str(e)}", exc_info=True)
        return None
    return plan


def update_plan(config, uuid, data):
    """
    Update a plan from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Updating plan {uuid}, {data}.")
    request = chartmogul.Plan.modify(config, uuid=uuid, data=data)
    try:
        plan = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error updating plan: {str(e)}", exc_info=True)
        return None
    return plan


## Plan groups Endpoints

def list_plan_groups(config, limit=20) -> list:
    """
    List all plan groups from ChartMogul API.

    Returns: A list of ChartMogul plan groups.
    """
    LOGGER.info(f"List plan groups.")
    all_plan_groups = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.PlanGroup.all(config, cursor=cursor, per_page=per_page)
        try:
            plan_groups = request.get()
            all_plan_groups.extend([parse_object(entry) for entry in plan_groups.plan_groups])
            total += per_page
            has_more = plan_groups.has_more
            cursor = plan_groups.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul plan groups: {str(e)}", exc_info=True)
            return None
    return all_plan_groups


def list_plan_group_plans(config, uuid, limit=20) -> list:
    """
    List all plans of a plan group from ChartMogul API.

    Returns: A list of ChartMogul plans of a plan group.
    """
    LOGGER.info(f"List plans of a plan group {uuid}.")
    all_plans = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.PlanGroup.all(config, uuid=uuid, cursor=cursor, per_page=per_page)
        try:
            plans = request.get()
            all_plans.extend([parse_object(entry) for entry in plans.plans])
            total += per_page
            has_more = plans.has_more
            cursor = plans.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul plans: {str(e)}", exc_info=True)
            return None
    return all_plans


def retrieve_plan_group(config, uuid):
    """
    Retrieve a plan group from ChartMogul API.

    Returns: The plan.
    """
    LOGGER.info(f"Retrieving plan group for {uuid}.")
    request = chartmogul.PlanGroup.retrieve(config, uuid=uuid)
    try:
        plan_group = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error retrieving plan group: {str(e)}", exc_info=True)
        return None
    return plan_group


def create_plan_group(config, data):
    """
    Create a plan group from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Creating plan group {data}.")
    request = chartmogul.PlanGroup.create(config, data=data)
    try:
        plan_group = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error creating plan group: {str(e)}", exc_info=True)
        return None
    return plan_group


def update_plan_group(config, uuid, data):
    """
    Update a plan group from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Updating plan group {uuid}, {data}.")
    request = chartmogul.PlanGroup.modify(config, uuid=uuid, data=data)
    try:
        plan_group = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error updating plan group: {str(e)}", exc_info=True)
        return None
    return plan_group



## Tasks Endpoints

def list_tasks(config, customer_uuid=None, assignee=None, due_date_on_or_after=None,
               estimated_close_date_on_or_before=None, completed=None, limit=20) -> list:
    """
    List all tasks from ChartMogul API.

    Returns: A list of ChartMogul tasks.
    """
    LOGGER.info(f"List tasks for {customer_uuid}, {assignee}, {due_date_on_or_after}, {estimated_close_date_on_or_before}, "
          f"{completed}.")
    all_tasks = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.Task.all(config,
                                      customer_uuid=customer_uuid,
                                      assignee=assignee,
                                      due_date_on_or_after=due_date_on_or_after,
                                      estimated_close_date_on_or_before=estimated_close_date_on_or_before,
                                      completed=completed,
                                      cursor=cursor,
                                      per_page=per_page)
        try:
            tasks = request.get()
            all_tasks.extend([parse_object(entry) for entry in tasks.entries])
            total += per_page
            has_more = tasks.has_more
            cursor = tasks.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul tasks: {str(e)}", exc_info=True)
            return None
    return all_tasks


def retrieve_task(config, uuid):
    """
    Retrieve a task from ChartMogul API.

    Returns: The task.
    """
    LOGGER.info(f"Retrieving task for {uuid}.")
    request = chartmogul.Task.retrieve(config, uuid=uuid)
    try:
        task = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error retrieving task: {str(e)}", exc_info=True)
        return None
    return task


def create_task(config, data):
    """
    Create a task from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Creating task {data}.")
    request = chartmogul.Task.create(config, data=data)
    try:
        task = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error creating task: {str(e)}", exc_info=True)
        return None
    return task


def update_task(config, uuid, data):
    """
    Update a task from ChartMogul API.

    Returns:
    """
    LOGGER.info(f"Updating task {uuid}, {data}.")
    request = chartmogul.Task.patch(config, uuid=uuid, data=data)
    try:
        task = parse_object(request.get())
    except Exception as e:
        LOGGER.error(f"Error updating task: {str(e)}", exc_info=True)
        return None
    return task


## Metrics API Endpoints

def all_metrics(config, start_date, end_date, interval, geo=None, plans=None) -> list:
    """
    List all metrics from ChartMogul API.

    Returns: A list of all metrics.
    """
    LOGGER.info(f"Fetching all metrics for {start_date}, {end_date}, {interval}, {geo}, {plans}.")
    request = chartmogul.Metrics.all(config,
                                     start_date=start_date,
                                     end_date=end_date,
                                     interval=interval,
                                     geo=geo,
                                     plans=plans
                                     )
    try:
        metrics = request.get()
        all_metrics = [parse_object(entry) for entry in metrics.entries]
    except Exception as e:
        LOGGER.error(f"Error fetching all metrics: {str(e)}", exc_info=True)
        return None
    return all_metrics


def mrr_metrics(config, start_date, end_date, interval, geo=None, plans=None) -> list:
    """
    List MRR metrics from ChartMogul API.

    Returns: A list of MRR metrics.
    """
    LOGGER.info(f"Fetching MRR metrics for {start_date}, {end_date}, {interval}, {geo}, {plans}.")
    request = chartmogul.Metrics.mrr(config,
                                     start_date=start_date,
                                     end_date=end_date,
                                     interval=interval,
                                     geo=geo,
                                     plans=plans
                                     )
    try:
        mrr = request.get()
        all_mrr = [parse_object(entry) for entry in mrr.entries]
    except Exception as e:
        LOGGER.error(f"Error fetching MRR metrics: {str(e)}", exc_info=True)
        return None
    return all_mrr


def arr_metrics(config, start_date, end_date, interval, geo=None, plans=None) -> list:
    """
    List ARR metrics from ChartMogul API.

    Returns: A list of ARR metrics.
    """
    LOGGER.info(f"Fetching ARR metrics for {start_date}, {end_date}, {interval}, {geo}, {plans}.")
    request = chartmogul.Metrics.arr(config,
                                     start_date=start_date,
                                     end_date=end_date,
                                     interval=interval,
                                     geo=geo,
                                     plans=plans
                                     )
    try:
        arr = request.get()
        all_arr = [parse_object(entry) for entry in arr.entries]
    except Exception as e:
        LOGGER.error(f"Error fetching ARR metrics: {str(e)}", exc_info=True)
        return None
    return all_arr


def arpa_metrics(config, start_date, end_date, interval, geo=None, plans=None) -> list:
    """
    List ARPA metrics from ChartMogul API.

    Returns: A list of ARPA metrics.
    """
    LOGGER.info(f"Fetching ARPA metrics for {start_date}, {end_date}, {interval}, {geo}, {plans}.")
    request = chartmogul.Metrics.arpa(config,
                                      start_date=start_date,
                                      end_date=end_date,
                                      interval=interval,
                                      geo=geo,
                                      plans=plans
                                      )
    try:
        arpa = request.get()
        all_arpa = [parse_object(entry) for entry in arpa.entries]
    except Exception as e:
        LOGGER.error(f"Error fetching ARPA metrics: {str(e)}", exc_info=True)
        return None
    return all_arpa


def asp_metrics(config, start_date, end_date, interval, geo=None, plans=None) -> list:
    """
    List ASP metrics from ChartMogul API.

    Returns: A list of ASP metrics.
    """
    LOGGER.info(f"Fetching ASP metrics for {start_date}, {end_date}, {interval}, {geo}, {plans}.")
    request = chartmogul.Metrics.asp(config,
                                     start_date=start_date,
                                     end_date=end_date,
                                     interval=interval,
                                     geo=geo,
                                     plans=plans
                                     )
    try:
        asp = request.get()
        all_asp = [parse_object(entry) for entry in asp.entries]
    except Exception as e:
        LOGGER.error(f"Error fetching ASP metrics: {str(e)}", exc_info=True)
        return None
    return all_asp


def customer_count_metrics(config, start_date, end_date, interval, geo=None, plans=None) -> list:
    """
    List Customer count metrics from ChartMogul API.

    Returns: A list of Customer count metrics.
    """
    LOGGER.info(f"Fetching Customer count metrics for {start_date}, {end_date}, {interval}, {geo}, {plans}.")
    request = chartmogul.Metrics.customer_count(config,
                                                start_date=start_date,
                                                end_date=end_date,
                                                interval=interval,
                                                geo=geo,
                                                plans=plans
                                                )
    try:
        customer_count = request.get()
        all_customer_count = [parse_object(entry) for entry in customer_count.entries]
    except Exception as e:
        LOGGER.error(f"Error fetching Customer count metrics: {str(e)}", exc_info=True)
        return None
    return all_customer_count


def customer_churn_rate_metrics(config, start_date, end_date, interval, geo=None, plans=None) -> list:
    """
    List Customer churn rate metrics from ChartMogul API.

    Returns: A list of Customer churn rate metrics.
    """
    LOGGER.info(f"Fetching Customer churn rate metrics for {start_date}, {end_date}, {interval}, {geo}, {plans}.")
    request = chartmogul.Metrics.customer_churn_rate(config,
                                                     start_date=start_date,
                                                     end_date=end_date,
                                                     interval=interval,
                                                     geo=geo,
                                                     plans=plans
                                                     )
    try:
        customer_churn_rate = request.get()
        all_customer_churn_rate = [parse_object(entry) for entry in customer_churn_rate.entries]
    except Exception as e:
        LOGGER.error(f"Error fetching Customer churn rate metrics: {str(e)}", exc_info=True)
        return None
    return all_customer_churn_rate


def mrr_churn_rate_metrics(config, start_date, end_date, interval, geo=None, plans=None) -> list:
    """
    List MRR churn rate metrics from ChartMogul API.

    Returns: A list of MRR churn rate metrics.
    """
    LOGGER.info(f"Fetching MRR churn rate metrics for {start_date}, {end_date}, {interval}, {geo}, {plans}.")
    request = chartmogul.Metrics.mrr_churn_rate(config,
                                                start_date=start_date,
                                                end_date=end_date,
                                                interval=interval,
                                                geo=geo,
                                                plans=plans
                                                )
    try:
        mrr_churn_rate = request.get()
        all_mrr_churn_rate = [parse_object(entry) for entry in mrr_churn_rate.entries]
    except Exception as e:
        LOGGER.error(f"Error fetching MRR churn rate metrics: {str(e)}", exc_info=True)
        return None
    return all_mrr_churn_rate


def ltv_metrics(config, start_date, end_date, interval, geo=None, plans=None) -> list:
    """
    List LTV metrics from ChartMogul API.

    Returns: A list of LTV metrics.
    """
    LOGGER.info(f"Fetching LTV metrics for {start_date}, {end_date}, {interval}, {geo}, {plans}.")
    request = chartmogul.Metrics.ltv(config,
                                     start_date=start_date,
                                     end_date=end_date,
                                     interval=interval,
                                     geo=geo,
                                     plans=plans
                                     )
    try:
        ltv = request.get()
        all_ltv = [parse_object(entry) for entry in ltv.entries]
    except Exception as e:
        LOGGER.error(f"Error fetching LTV metrics: {str(e)}", exc_info=True)
        return None
    return all_ltv


## Subscription Events

def list_subscription_events(config, data_source_uuid=None, external_id=None, customer_external_id=None,
                             subscription_external_id=None, event_type=None, event_date=None, effective_date=None,
                             plan_external_id=None, limit=20) -> list:
    """
    List all subscription events from ChartMogul API.

    Returns: A list of ChartMogul subscription events.
    """
    LOGGER.info(f"List subscription events for {data_source_uuid}, {external_id}, {customer_external_id}, {event_type}, "
                f"{subscription_external_id}, {event_date}, {effective_date}, {plan_external_id}.")
    all_subscription_events = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.SubscriptionEvent.all(config,
                                                   data_source_uuid=data_source_uuid,
                                                   external_id=external_id,
                                                   customer_external_id=customer_external_id,
                                                   subscription_external_id=subscription_external_id,
                                                   event_type=event_type,
                                                   event_date=event_date,
                                                   effective_date=effective_date,
                                                   plan_external_id=plan_external_id,
                                                   cursor=cursor,
                                                   per_page=per_page)
        try:
            subscription_events = request.get()
            all_subscription_events.extend([parse_object(entry) for entry in subscription_events.subscription_events])
            total += per_page
            has_more = subscription_events.has_more
            cursor = subscription_events.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul subscription events: {str(e)}", exc_info=True)
            return None
    return all_subscription_events


## Invoices

def list_invoices(config, data_source_uuid=None, external_id=None, customer_uuid=None,
                  validation_type=None, limit=20) -> list:
    """
    List all invoices from ChartMogul API.

    Returns: A list of ChartMogul invoices.
    """
    LOGGER.info(f"List invoices for {data_source_uuid}, {external_id}, {customer_uuid}, {validation_type}.")
    all_invoices = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.Invoice.all(config,
                                         data_source_uuid=data_source_uuid,
                                         external_id=external_id,
                                         customer_uuid=customer_uuid,
                                         validation_type=validation_type,
                                         cursor=cursor,
                                         per_page=per_page)
        try:
            invoices = request.get()
            all_invoices.extend([parse_object(entry) for entry in invoices.invoices])
            total += per_page
            has_more = invoices.has_more
            cursor = invoices.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul invoices: {str(e)}", exc_info=True)
            return None
    return all_invoices


## Activities

def list_activities(config, start_date=None, end_date=None, type=None, order=None, limit=20) -> list:
    """
    List all activities from ChartMogul API.

    Returns: A list of ChartMogul activities.
    """
    LOGGER.info(f"List activities for {start_date}, {end_date}, {type}, {order}.")
    all_activities = []
    has_more = True
    cursor = None
    per_page = 20
    total = 0
    while has_more and total < limit:
        request = chartmogul.Activity.all(config,
                                         start_date=start_date,
                                         end_date=end_date,
                                         type=type,
                                         order=order,
                                         cursor=cursor,
                                         per_page=per_page)
        try:
            activities = request.get()
            all_activities.extend([parse_object(entry) for entry in activities.entries])
            total += per_page
            has_more = activities.has_more
            cursor = activities.cursor
        except Exception as e:
            LOGGER.error(f"Error fetching ChartMogul activities: {str(e)}", exc_info=True)
            return None
    return all_activities


def parse_object(obj):
    if isinstance(obj, datetime.datetime) or isinstance(obj, datetime.date):
        return obj.isoformat() 
    elif hasattr(obj, '__dict__'):
        result = {}
        for key, value in obj.__dict__.items():
            result[key] = parse_object(value)
        return result
    elif isinstance(obj, list):
        return [parse_object(item) for item in obj]
    else:
        return obj